/*
 * NS2NP - Converts Netscape Bookmarks 
 *         to NetPositive-"Bookmarks"
 *         i.e. creates the corresponding directories and files
 *
 * small, but useful utility :)
 * (C)opyleft 1998 by Rosso
 * <http://www.bigfoot.com/~mcrosso>
 *
 * Version HISTORY:
 *
 * Date		Version		Comments
 * 
 * 13.1.99	1.0.4		+ BUG found: wrong parameter-types for File->Read (now using ssize_t !)
 						+ Summary with total number of urls and dirs is now displayed at the end
 						+ if the destination-path does not exist, the directory will be created
 *						+ First public release (my FIRST BeWare !) :-)
 * 12.1.99  1.0.2		+ argument-checking, brushed up source, small documentation
 						# BUG: crashes the first time it is run, but works ?!
 * 10.1.99	1.0.0		+ dir-structure build correctly, parser-bug found (URL/TOKEN was too small),
 						+ "/" is recognised as forbidden character and replaced by a "\"
 * 3.1.99	0.0.5		+ engine works, urls parsed correctly
 						# dir-structure wrong, parser gets quite crazy after some 50% of my Bookmarks-
 						  file (ca. 700 Links)
 */

#include <stdio.h>
#include <stddef.h>
#include <string.h>

#include <storage/Directory.h>
#include <storage/Entry.h>
#include <storage/File.h>
#include <storage/Node.h>
#include <storage/Path.h>

#define LEN_BUF   1024
#define LEN_TOKEN 127	// must be able to hold long URLs like in <A HREF="...">
#define LEN_DATA  127	// lets stay on the safe side here, too

#ifndef true 
	#define true  1
	#define false 0
#endif

// TODO:+ File-Check: Is a Netscape-bookmark-file ? (-> else: warning)
//		+ ExploDer-Bookmarks ? -> on request
//		+ make this a real BeApp with a nice GUI ...!

int strmod(char source[LEN_DATA], char filter, char mod)
{
	int  c = 0, mod_count = 0;
	
	while (source[c] != '\0')
	{
		if (source[c] == filter)
		{
			source[c] = mod;
			mod_count++;
		}
		
		c++;
	}
	return mod_count;
}

int main(int argc, char **argv)
{
	const char	*VERSION = "1.0.4";
	const char	bookmark[B_ATTR_NAME_LENGTH] = "application/x-vnd.Be-bookmark";
	const char	pref_app[B_ATTR_NAME_LENGTH] = "application/x-vnd.Be-NPOS";

	BFile		*fileBookmarks;		// Input and
	BFile		*fileNewBookmarks;	// output file(s)
	BPath		*pathBookmarks;
	BDirectory	*dirBookmarks;
	BNode		*bookNode = NULL;
	ssize_t		buf_read = 0;
	char		buffer[LEN_BUF], token[LEN_TOKEN], data[LEN_DATA];
	enum		mode {MOD_COMMENT = 0, MOD_TOKEN, MOD_DATA, MOD_PROCESS};				// parse-modes
	uint		arg, ch, mode = MOD_COMMENT, pos_buf = 0, pos_token = 0, pos_data = 0,	// initialise parser
				total_urls = 0, total_dirs = 0;
	char		*c = NULL, *url = NULL;
	bool		VERBOSE = false;
	status_t 	err;
	
	// BApplication app("application/x-whatever");
	
	// ********* Handle Args
	
	if ( (argc == 1) || ((argc == 2) && (( (strcmp(argv[1], "-h") == 0) || (strcmp(argv[1], "-v") == 0)))))
	{	// print Usage
		printf("%s %s (C)opyright 1999 by Gregor B. Rosenauer (mcrosso@bigfoot.com)\n", argv[0], VERSION);
		printf("JoyWare - Use and Enjoy ;) And if you like it, please send me a note :-)\n\n");
		
		printf("Usage: %s [-h|v] <Netscape-Bookmark-File> [Be-Bookmark-Path]\n\n", argv[0]);
		
		printf("-v  turns on verbose mode (shows files and directories created)\n");
		printf("-h  shows this help\n");
		printf("The 2nd parameter specifies where your Bookmarkfile is located.\n");
		printf("The 3rd parameter specifies the root of the Be-Bookmark-directory.\n");
		printf("    It defaults to NetPositive's settings-directory and you should not\n");
		printf("    need to change it unless you use another browser.\n\n");
		
		printf("Note 1: Currently only Netscape Bookmarkfiles are supported.\n");
		printf("Note 2: Bookmark-files are OVERWRITTEN and updated, so please back up\n");
		printf("        your old bookmarkfiles to go sure nothing unwanted happens.\n");
		printf("        NetPositive's Bookmark-directory normally is:\n");
		printf("        \"/boot/home/config/settings/NetPositive/Bookmarks/\"\n");
		return 1;
	}
	
	arg = 1; argc--;		// arg = current argument to be checked

	if (strcmp(argv[arg], "-v") == 0)
	{
		VERBOSE = true;
		arg++; argc--;
	}

	// open input file
	fileBookmarks = new BFile(argv[arg], B_READ_ONLY);
	
	if ((err = fileBookmarks->InitCheck()) != B_NO_ERROR)
	{
		fprintf(stderr, "%s: Error opening file %s: %s\n", argv[0], argv[arg], strerror(err));
		return 1;
	}

	arg++; argc--;
	
	if (argc == 0)		// no path specified -> use default path
		pathBookmarks = new BPath("/boot/home/config/settings/NetPositive/Bookmarks");
	else
		pathBookmarks = new BPath(argv[arg]);		// use user-path
		
	dirBookmarks  = new BDirectory(pathBookmarks->Path());

	err = dirBookmarks->CreateDirectory(pathBookmarks->Path(), dirBookmarks);
	if (err != B_OK)
	{
		if (err != B_FILE_EXISTS)
		{
			fprintf(stderr, "%s: Error creating directory \"%s\": %s !\n", argv[0], data, strerror(err));
			return 1;
		}
	}
	else printf("[Destination-Directory created.]\n");


	// ********* Parse bookmark-file	

	token[0] = '\0'; data[0] = '\0';
	
 	// read as long as there is data...
	while ((buf_read = fileBookmarks->Read((void *)buffer, LEN_BUF)) > 0)
	{
		//printf("Reading from file...\n");	// DB

		for (pos_buf = 0; pos_buf < LEN_BUF; pos_buf++)
		{
			// read next char
			ch = buffer[pos_buf];

			// '<' - beginning of token ?
			if (ch == '<')
			{
				if (mode == MOD_DATA)
					mode = MOD_PROCESS;
				else
					mode = MOD_TOKEN;
			}
			else if (ch == '>')
				{
					if (mode != MOD_COMMENT) mode = MOD_DATA;
				}
			else
			{
				switch (mode)
				{				
					case MOD_TOKEN:
						if (pos_token == 3)			// check for comment
						{
							if (token[0]=='!' && token[1]=='-' && token[2]=='-')
							{
								mode = MOD_COMMENT;			// ignore rest of comment
								pos_token = pos_data = 0;
								token[0] = data[0] = '\0';
							}
						}
						if (mode != MOD_COMMENT)
						{
							token[pos_token++] = ch;
						}
						
						break;
						
					case MOD_DATA:
						data [pos_data++]  = ch;
						
						break;
						
					case MOD_PROCESS:
						// data should not be null (for files/dirs)
						if (pos_data == 0) data[pos_data++] = ' ';
						
						token[pos_token] = '\0';
						data [pos_data]  = '\0';

						if ( strcmp(token, "/H3") != 0)
						{	if (strstr(token, "H3") != 0)				// new tree
							{
								(void) strmod(data, '/', '\\');			// '/' in the name would specify a directory and must be changed !
						
								pathBookmarks->Append(data);
								
								if (VERBOSE)
								{
									printf("Creating DIR \"%s\"\n", pathBookmarks->Path());	// DB
								}
								delete(dirBookmarks);
								dirBookmarks = new BDirectory(pathBookmarks->Path());

								err = dirBookmarks->CreateDirectory(pathBookmarks->Path(), dirBookmarks);
								if ((err != B_OK) && (err != B_FILE_EXISTS))
								{
									fprintf(stderr, "%s: Error creating directory \"%s\": %s !\n", argv[0], data, strerror(err));
								}
								else total_dirs++;
							}
						}
						if ( strcmp(token, "/DL") == 0)				// End of (Sub-)Tree
						{
							pathBookmarks->GetParent(pathBookmarks);
///							printf("--- Parent Path: %s\n", pathBookmarks->Path());			// DB
						}
						else if ( (c = strstr(token, "A HREF")) != NULL)
						{
							if (  (c = strstr(token, "\"")) != NULL)
							{
								int counter = 0;

								url = c;
								
								while ((url[counter] = *++c) != '\"')
									counter++;
								url[counter] = '\0';
								
								// build new (absolute) filename
								(void) strmod(data, '/', '\\');		// '/' in the name would specify a directory and must be changed !
//								printf("-> Normalized Path is : %s\n", data);	// DB
								
								pathBookmarks->Append(data);

								if (VERBOSE)
								{
									printf("\tAdding Bookmark \"%s\"\n", data);
								}
								fileNewBookmarks = new BFile(pathBookmarks->Path(), B_READ_WRITE);

								err = dirBookmarks->CreateFile(pathBookmarks->Path(), fileNewBookmarks, false);
								if (err != B_NO_ERROR)
								{
									/* The Bookmark-file could not be created */
									fprintf(stderr, "%s: File \"%s\" could not be created: %s !\n", argv[0], data, strerror(err));
								}
								else
								{
									// transform the file into a node so we can write the attributes									
									bookNode = new BNode(pathBookmarks->Path());
									
									// write attributes
									// this might look a bit hacky but I doubt NetPositive does it differently ;)
									bookNode->WriteAttr("BEOS:TYPE",     B_STRING_TYPE, 0, bookmark, B_ATTR_NAME_LENGTH);
									bookNode->WriteAttr("BEOS:PREF_APP", B_STRING_TYPE, 0, pref_app, B_ATTR_NAME_LENGTH);
									bookNode->WriteAttr("META:url",      B_STRING_TYPE, 0, url, B_ATTR_NAME_LENGTH);
									bookNode->WriteAttr("META:title",    B_STRING_TYPE, 0, data, B_ATTR_NAME_LENGTH);

									delete(bookNode);
									total_urls++;
								}

								delete(fileNewBookmarks);
								pathBookmarks->GetParent(pathBookmarks);  // get "parent" from file!
///								printf("\t  Path (Parent):%s\n", pathBookmarks->Path());	// DB
							}
							else
								fprintf(stderr, "%s: Malformed Link !\n", argv[0]);
						}
							
						pos_token = pos_data = 0;
						data [pos_data]  = '\0';

						mode = MOD_TOKEN;
						token[pos_token++] = ch;
						
						break;
						
					case MOD_COMMENT:
						
						break;
				}
			}
		}
	}

	if (pathBookmarks != NULL) delete (pathBookmarks);
	if (dirBookmarks  != NULL) delete (dirBookmarks);
	if (bookNode      != NULL) delete (bookNode);

	if (VERBOSE) printf("\nSummary: %d dirs and %d urls created.\n", total_dirs, total_urls);	
	return 0;
}
